/*
* Copyright (C) 2021, KylinSoft Co., Ltd.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 3, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, see <http://www.gnu.org/licenses/&gt;.
*
*/
#include "scansettingswidget.h"
#include "include/common.h"

#include <ukui-log4qt.h>

#include <QFileInfo>
#include <QFileDialog>
#include <QMessageBox>
#include <QWidgetList>
#include <QApplication>
#include <QListView>
#include <QStandardItem>
#include <QStandardItemModel>
#include <QDateTime>
#include <buried_point.hpp>
#include <gsettings.hpp>
#include "include/theme.h"

ScanSettingsWidget::ScanSettingsWidget(QWidget *parent) :
    QWidget(parent),
    scanButtonLeftLabel(new QLabel()),
    scanButtonRightLabel(new QLabel()),
    scanButtonHLayout(new QHBoxLayout()),
    m_scanButton(new QPushButton()),
    m_deviceSettingsLabel(new QLabel()),
    m_deviceLabel(new QLabel()),
    m_deviceComboBox(new QComboBox()),
    m_pageNumberLabel(new QLabel()),
    m_pageNumberComboBox(new QComboBox()),
    m_timeLabel(new QLabel()),
    m_timeComboBox(new QComboBox()),
    m_typeLabel(new QLabel()),
    m_typeComboBox(new QComboBox()),
    m_colorLabel(new QLabel()),
    m_colorComboBox(new QComboBox()),
    m_resolutionLabel(new QLabel()),
    m_resolutionComboBox(new QComboBox()),
    m_fileSettingsLabel(new QLabel()),
    m_sizeLabel(new QLabel()),
    m_sizeComboBox(new QComboBox()),
    m_formatLabel(new QLabel()),
    m_formatComboBox(new QComboBox()),
    m_saveNameLabel(new QLabel()),
    m_saveNameEdit(new QLineEdit()),
    m_saveDirectoryLabel(new QLabel()),
    m_saveDirectoryButtonLabel(new QLabel()),
    m_saveDirectoryButtonLayout(new QHBoxLayout),
    m_saveDirectoryButton(new QPushButton()),
    m_settingsFormLayout(new QFormLayout()),
    m_sendMailButton(new QPushButton()),
    m_SaveAsButton(new QPushButton()),
    m_buttonsHLayout(new QHBoxLayout()),
    m_mainScrollArea (new QScrollArea()),
    m_mainVLayout(new QVBoxLayout(this))
{
    setupGui();
//    initTheme();
    initConnect();
}
void ScanSettingsWidget::initTheme(){
    QPalette pal(palette());
    if (isDarkTheme()) {
        pal.setColor(QPalette::Background, QColor(18, 18, 18));
    } else {
        pal.setColor(QPalette::Background, QColor(255, 255, 255));
    }
    setAutoFillBackground(true);
    setPalette(pal);
}
void ScanSettingsWidget::paintEvent(QPaintEvent *event)
{
    Q_UNUSED(event);
    QPainter p(this);
    p.setRenderHint(QPainter::Antialiasing);
    QPainterPath rectPath;
    rectPath.addRoundedRect(this->rect(), 0, 0);
    QStyleOption opt;
    opt.init(this);

    QColor mainColor;
    mainColor = opt.palette.color(QPalette::Base);

    p.fillPath(rectPath,QBrush(mainColor));
}

void ScanSettingsWidget::selectSaveDirectorySlot()
{
    if (currentSaveDirectory.isEmpty()) {
        QString scannerPicturePath = g_config_signal->m_scannerImagePath;
        currentSaveDirectory = scannerPicturePath ;
    }

    QString midPath = currentSaveDirectory;
    QString dlgTitle = tr("Select a directory");
    QString selectedDir = QFileDialog::getExistingDirectory(this, dlgTitle, currentSaveDirectory, QFileDialog::ShowDirsOnly);
    KyInfo() << "selected directory: " << selectedDir;

    if (!selectedDir.isEmpty()) {
        QFileInfo file(selectedDir);
        if (file.permission(QFileDevice::WriteUser | QFileDevice::ReadGroup)) {
            KyInfo() << "The user could read and write " << selectedDir;

            currentSaveDirectory = selectedDir;
            setSaveButtonLabelAttributes(m_saveDirectoryButtonLabel, currentSaveDirectory, ScanSettingsButtonElideWidth);
        } else {
            if (file.isWritable()) {
                KyInfo() << "The user can read and write " << selectedDir;

                currentSaveDirectory = selectedDir;
                setSaveButtonLabelAttributes(m_saveDirectoryButtonLabel, currentSaveDirectory, ScanSettingsButtonElideWidth);

                return;
            }
            KyInfo() << "The user can't read and write " << selectedDir;

            QString msg = tr("Currently user has no permission to modify directory ") + selectedDir;
            warnMsg(msg);
        }
    }
    g_sane_object->userInfo.saveDirectory = currentSaveDirectory;

    KyInfo() << "saveDirectory = " << currentSaveAsDirectory;
}

void ScanSettingsWidget::deviceCurrentTextChangedSlot(QString text)
{
    KyInfo() << "Sane device name: " << g_sane_object->userInfo.name << "current devece name: " << m_deviceComboBox->currentText();

    if (g_sane_object->userInfo.name.isEmpty()) {
        g_sane_object->userInfo.name = text;

        KyInfo() << "We do not need open device again.";
        return;
    } else {
        KyInfo() << "Device changed, therefore, we need open device again.";
        g_sane_object->userInfo.name = text;
    }

    int curTextLen = m_deviceComboBox->currentText().length();

    if ( curTextLen >= 20) {
        m_deviceComboBox->setToolTip(m_deviceComboBox->currentText());
    } else {
        m_deviceComboBox->setToolTip("");
    }

    // index is chosen by user manually, or default(0)
    int index = m_deviceComboBox->currentIndex();
    if (index == -1) {
        index = 0;
    }

    g_sane_object->userInfo.deviceNameIndex = index;

    g_sane_object->saneClose();

    // while switch scan device, we should open the scan device to get some parameters
    g_sane_object->openDeviceIndex = index;
    g_sane_object->openSaneDeviceForPage(index);

    updateSettingsForSwitchDevices();
}

/**
 * @brief ScanSettingsWidget::pageNumberCurrentTextChangedSlot
 * @param text
 *
 */
void ScanSettingsWidget::pageNumberCurrentTextChangedSlot(QString text)
{
    if (text.compare("Multiple", Qt::CaseInsensitive) == 0
            || text.compare("多页扫描", Qt::CaseInsensitive) == 0) {
        g_sane_object->userInfo.pageNumber = tr("Multiple");
        m_saveNameEdit->setMaxLength(240);

        // Avoid SANE_STATUS_NO_DOC Error to set time not enable.
        g_sane_object->setSaneStatus(true);
        showTimeRow();
        kdk::kabase::BuriedPoint buriedPointTest;
        if (buriedPointTest.functionBuriedPoint(kdk::kabase::AppName::KylinScanner , kdk::kabase::BuriedPoint::PT::KylinScannerMultiPageScan)) {
            qCritical() << "Error : buried point fail !";
        };
    } else {
        g_sane_object->userInfo.pageNumber = tr("Single");
        m_saveNameEdit->setMaxLength(246);
        hideTimeRow();
        kdk::kabase::BuriedPoint buriedPointTest;
        if (buriedPointTest.functionBuriedPoint(kdk::kabase::AppName::KylinScanner , kdk::kabase::BuriedPoint::PT::KylinScannerSinglePageScan)) {
            qCritical() << "Error : buried point fail !";
        };
    }
}

void ScanSettingsWidget::timeCurrentTextChangedSlot(QString text)
{
    g_sane_object->userInfo.time = text;

    KyInfo() << "userInfo.time = " << text;
}

void ScanSettingsWidget::typeCurrentTextChangedSlot(QString text)
{
//    if (0 == QString::compare(text, tr("ADF Back"), Qt::CaseInsensitive)) {
//        g_sane_object->userInfo.type = "ADF Back";
//    } else if(0 == QString::compare(text, tr("ADF Front"), Qt::CaseInsensitive)) {
//        g_sane_object->userInfo.type = "ADF Front";
//    } else if(0 == QString::compare(text, tr("ADF Duplex"), Qt::CaseInsensitive)) {
//        g_sane_object->userInfo.type = "ADF Duplex";
//    }
    g_sane_object->userInfo.type = text;

    KyInfo() << "userInfo.type = " << text;
}

void ScanSettingsWidget::colorCurrentTextChangedSlot(QString text)
{
    if (0 == QString::compare(text, tr("Color"), Qt::CaseInsensitive)) {
        g_sane_object->userInfo.color = "Color";
    } else if(0 == QString::compare(text, tr("Lineart"), Qt::CaseInsensitive)) {
        g_sane_object->userInfo.color = "Lineart";
    } else {
        g_sane_object->userInfo.color = "Gray";
    }

    KyInfo() << "userInfo.color = " << text;
}

void ScanSettingsWidget::resolutionCurrentTextChangedSlot(QString text)
{
    if ((0 == text.compare(tr("4800 dpi"), Qt::CaseInsensitive))
         || (0 == text.compare(tr("2400 dpi"), Qt::CaseInsensitive))
         || (0 == text.compare(tr("1200 dpi"), Qt::CaseInsensitive))) {

        QString msg = tr("This resolution will take a loog time to scan, please choose carelly.");
        warnMsg(msg);
    }

    g_sane_object->userInfo.resolution = text;

    KyInfo() << "userInfo.resolution = " << text;
}

void ScanSettingsWidget::sizeCurrentTextChangedSlot(QString text)
{
    g_sane_object->userInfo.size = text;
    KyInfo() << "userInfo.size = " << text;
}

void ScanSettingsWidget::formatCurrentTextChangedSlot(QString text)
{
    g_sane_object->userInfo.format = text;
    KyInfo() << "userInfo.format = " << text;
}

void ScanSettingsWidget::nameCurrentTextChangedSlot(QString text)
{
    QString msg;
    if (m_saveNameEdit->text().contains(QChar('/'), Qt::CaseInsensitive)) {
        msg = tr("cannot contain '/' character.");
        warnMsg(msg);
        m_saveNameEdit->cursorBackward(true, 1); // move left a character
        m_saveNameEdit->del(); // delete right or selected text
    }
    if (m_saveNameEdit->text().startsWith(QChar('.'), Qt::CaseInsensitive)) {
        msg = tr("cannot save as hidden file.");
        warnMsg(msg);
        m_saveNameEdit->cursorBackward(true, 1);
        m_saveNameEdit->del();
    }
    QString saveNameEditStr = m_saveNameEdit->text();
    int len = strlen(saveNameEditStr.toLocal8Bit());
    qDebug() << "name length:"<<len;

    g_sane_object->userInfo.saveName = m_saveNameEdit->text();
    KyInfo() << "saveName = " << g_sane_object->userInfo.saveName;

    setNameEditTooltip();
}

void ScanSettingsWidget::sendMailButtonClickedSlot()
{
    kdk::kabase::BuriedPoint buriedPointTest;
    if (buriedPointTest.functionBuriedPoint(kdk::kabase::AppName::KylinScanner , kdk::kabase::BuriedPoint::PT::KylinScannerSendMail)) {
        qCritical() << "Error : buried point fail !";
    };
    int retDialog;

    AppList *maillist = getAppIdList(MailType);

    if (! maillist) {
        NoMailDialog *dialog = new NoMailDialog();

        retDialog = dialog->exec();
        if (retDialog == QDialog::Accepted) {
            QProcess *process = new QProcess();
            process->start("/usr/bin/kylin-software-center");
        }
        delete dialog;
    } else {
        KyInfo() << "Get mail client list success.";

        SendMailDialog *dialog = new SendMailDialog(this);

        dialog->setMailSelectComboboxItems();

        dialog->exec();
    }

    g_user_signal->sendMailButtonClicked();
}

void ScanSettingsWidget::saveAsButtonClickedSlot()
{
    kdk::kabase::BuriedPoint buriedPointTest;
    if (buriedPointTest.functionBuriedPoint(kdk::kabase::AppName::KylinScanner , kdk::kabase::BuriedPoint::PT::KylinScannerSaveAs)) {
        qCritical() << "Error : buried point fail !";
    };
    QString filename;
    QString filepath;
    QString filetype;
    QString msg;
    QString titlename;
    bool flagSave = false;

    while (! flagSave) {
        if (g_sane_object->ocrFlag == 1) {
            if (m_saveNameEdit->text().endsWith(".txt")) {
                filename = m_saveNameEdit->text();
            } else {
                filename = m_saveNameEdit->text() + ".txt";
            }
            titlename = tr("Store text");
        } else {
            if (m_saveNameEdit->text().endsWith(".jpg")
                    || m_saveNameEdit->text().endsWith(".png")
                    || m_saveNameEdit->text().endsWith(".bmp")
                    || m_saveNameEdit->text().endsWith(".pdf")) {
                filename = m_saveNameEdit->text();
            } else {
                filename = m_saveNameEdit->text() + "." + m_formatComboBox->currentText();
            }
            titlename = tr("Save as dialog");
        }
        filepath = currentSaveDirectory;

        KyInfo() << "filepath = " << filepath;


        QWidget *widget = nullptr;
        QWidgetList widgetList = QApplication::allWidgets();
        for (int i=0; i<widgetList.length(); ++i) {
            if (widgetList.at(i)->objectName() == "MainWindow") {
                widget = widgetList.at(i);
            }
        }

        SaveFileDialog *saveDialog = new SaveFileDialog(widget, g_sane_object->ocrFlag, filename, titlename);

        if (widget) {
            QRect rect = widget->geometry();
            int x = rect.x() + rect.width()/2 - saveDialog->width()/2;
            int y = rect.y() + rect.height()/2 - saveDialog->height()/2;
            saveDialog->move(x,y);
        }

        saveDialog->setSaveAsDirectory(filepath);

        if (saveDialog->exec() == QFileDialog::Accepted) {
            filepath = saveDialog->selectedFiles().at(0);
            filename = saveDialog->getFileName();
            KyInfo() << "filepath = " << filepath
                     << "filename = " << filename;

            if (filepath.isNull())
                break;

            if (filename.contains(QChar('/'), Qt::CaseInsensitive)) {
                msg = tr("cannot contain '/' character.");
                warnMsg(msg);
                continue;
            }
            if (filename.startsWith(QChar('.'), Qt::CaseInsensitive)) {
                msg = tr("cannot save as hidden file.");
                warnMsg(msg);
                continue;
            }

            // While not endsWith (".jpg" | ".png" | "pdf" | "bmp"),
            // we neet add suffix behind.
            if (g_sane_object->ocrFlag == 1) {
                if (!filename.endsWith(".txt", Qt::CaseInsensitive)) {
                    filetype = saveDialog->getFileType();
                    filepath = filepath.append(filetype);
                }
            } else {
                if (!filename.endsWith(".jpg", Qt::CaseInsensitive)
                        && !filename.endsWith(".png", Qt::CaseInsensitive)
                        && !filename.endsWith(".pdf", Qt::CaseInsensitive)
                        && !filename.endsWith(".bmp", Qt::CaseInsensitive)) {

                    filetype = saveDialog->getFileType();
                    filepath = filepath.append(filetype);
                }
            }
            flagSave = true;
        } else {
            flagSave = false;
            return;
        }
    }

    KyInfo() << "filepath = " << filepath;

    if (flagSave) {
        g_user_signal->exitWindowWithSaveFlag = false;
        g_user_signal->saveAsButtonClicked(filepath);
    }
}

void ScanSettingsWidget::scanButtonClickedSlot()
{
    updateSettingsStatusForStartScan();

    g_user_signal->startScanOperation();
}

void ScanSettingsWidget::fontSizeChanged()
{
    setLabelAttributes(m_deviceLabel, tr("Device"));
    setLabelAttributes(m_pageNumberLabel, tr("Pages"));
    setLabelAttributes(m_timeLabel, tr("Time"));
    setLabelAttributes(m_typeLabel, tr("Type"));
    setLabelAttributes(m_colorLabel, tr("Colour"));
    setLabelAttributes(m_resolutionLabel, tr("Resolution"));
    setLabelAttributes(m_sizeLabel, tr("Size"));
    setLabelAttributes(m_formatLabel, tr("Format"));
    setLabelAttributes(m_saveNameLabel, tr("Name"));
    setLabelAttributes(m_saveDirectoryLabel, tr("Save"));

    setSaveButtonLabelAttributes(m_saveDirectoryButtonLabel, currentSaveDirectory, ScanSettingsButtonElideWidth);

    m_sendMailButton->adjustSize();
    m_sendMailButton->setText(tr("Mail to"));

    m_SaveAsButton->adjustSize();
    m_SaveAsButton->setText(currentSaveAsDirectory);

}

void ScanSettingsWidget::fontSizeChangedSlot()
{
    fontSizeChanged();
    int systemFontSize = kdk::kabase::Gsettings::getSystemFontSize().toInt();
    QFont font;
    font.setPointSize(systemFontSize);
//    m_scanButton->setFont(font);
    scanButtonRightLabel->setFont(font);

}
bool ScanSettingsWidget::isDarkTheme()
{
    QString systemTheme = kdk::kabase::Gsettings::getSystemTheme().toString();

    if (systemTheme == STYLE_NAME_KEY_DARK || systemTheme == STYLE_NAME_KEY_BLACK) {
        return true;
    } else {
        return false;
    }
}
void ScanSettingsWidget::setupGui()
{
    this->setProperty("useSystemStyleBlur", true);
    this->setAttribute(Qt::WA_TranslucentBackground, true);
    this->setFixedWidth(ScanSettingsWidgetWidth);

    scanButtonLeftLabel->setPixmap(QPixmap(":/default-connect-page/begin-scan.svg"));
    scanButtonRightLabel->setText(tr("Begin Scan"));
//    scanButtonRightLabel->setScaledContents(true);

    scanButtonHLayout->addStretch();
    scanButtonHLayout->addWidget(scanButtonLeftLabel);
    scanButtonHLayout->addSpacing(4);
    scanButtonHLayout->addWidget(scanButtonRightLabel);
    scanButtonHLayout->addStretch();
    scanButtonHLayout->setSpacing(4);
    scanButtonHLayout->setContentsMargins(4, 0, 0, 0);

    m_scanButton->setLayout(scanButtonHLayout);
//    m_scanButton->setFixedSize(QSize(193+AddWidthForLargeFontSize, 36));
    m_scanButton->setCursor(Qt::PointingHandCursor);
    scanButtonRightLabel->setStyleSheet("color:white;");
//    m_scanButton->setStyleSheet("QPushButton{background: rgb(55, 144, 250);}");
    QPalette palette;
    QColor color(55, 144, 250);
    palette.setColor(QPalette::Button, color);
    m_scanButton->setPalette(palette);

//    m_deviceComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

    m_deviceSettingsLabel->setText(tr("Scanner device"));
    QPalette palFont;
    palFont.setColor(QPalette::WindowText,QColor("#8C8C8C"));
    m_deviceSettingsLabel->setPalette(palFont);

//    m_pageNumberComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

//    m_timeComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

//    m_typeComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

//    m_colorComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

//    m_resolutionComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

    m_fileSettingsLabel->setText(tr("File settings"));
    m_fileSettingsLabel->setPalette(palFont);

//    m_sizeComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

//    m_formatComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);

    m_saveNameEdit->setText(tr("scanner01"));
    m_saveNameEdit->setMaxLength(246);
//    m_saveNameEdit->setFixedSize(ScanSettingsWidgetComboboxSize);

    m_saveDirectoryButtonLayout->addWidget(m_saveDirectoryButtonLabel);
    m_saveDirectoryButtonLayout->setSpacing(0);
    m_saveDirectoryButtonLayout->setContentsMargins(8, 0, 0, 0);
    m_saveDirectoryButton->setLayout(m_saveDirectoryButtonLayout);
//    m_saveDirectoryButton->setFixedSize(ScanSettingsWidgetComboboxSize);

    if (currentSaveDirectory.isEmpty()) {
        QString scannerPicturePath = g_config_signal->m_scannerImagePath;
        currentSaveDirectory = scannerPicturePath ;

        g_sane_object->userInfo.saveDirectory = currentSaveDirectory;

        KyInfo() << "saveDirectory = " << currentSaveDirectory;
    }

    currentSaveAsDirectory = tr("Save as");
    fontSizeChanged();

    int spacing = 7;


    m_settingsFormLayout->setSpacing(0);
    m_settingsFormLayout->setHorizontalSpacing(spacing);
    m_settingsFormLayout->setVerticalSpacing(8);
    m_settingsFormLayout->addRow(m_scanButton);
    m_settingsFormLayout->addRow(m_deviceSettingsLabel);
    m_settingsFormLayout->addRow(m_deviceLabel, m_deviceComboBox);
    m_settingsFormLayout->addRow(m_pageNumberLabel, m_pageNumberComboBox);
    m_settingsFormLayout->addRow(m_typeLabel, m_typeComboBox);
    m_settingsFormLayout->addRow(m_colorLabel, m_colorComboBox);
    m_settingsFormLayout->addRow(m_resolutionLabel, m_resolutionComboBox);
    m_settingsFormLayout->addRow(m_fileSettingsLabel);
    m_settingsFormLayout->addRow(m_sizeLabel, m_sizeComboBox);
    m_settingsFormLayout->addRow(m_formatLabel, m_formatComboBox);
    m_settingsFormLayout->addRow(m_saveNameLabel, m_saveNameEdit);
    m_settingsFormLayout->addRow(m_saveDirectoryLabel, m_saveDirectoryButton);
    m_settingsFormLayout->setLabelAlignment(Qt::AlignRight | Qt::AlignVCenter);
    m_settingsFormLayout->setFormAlignment(Qt::AlignLeft );
    m_settingsFormLayout->setContentsMargins(24, 0, 24, 0);

    m_sendMailButton->setMinimumSize(ScanSettingsWidgetButtonSize);
    m_SaveAsButton->setMinimumSize(ScanSettingsWidgetButtonSize);

    m_buttonsHLayout->setSpacing(0);
    //m_buttonsHLayout->addSpacing(16-spacing);
    m_buttonsHLayout->addWidget(m_sendMailButton);
    m_buttonsHLayout->addSpacing(12);
    //m_buttonsHLayout->addStretch();
    m_buttonsHLayout->addWidget(m_SaveAsButton);
    //m_buttonsHLayout->addSpacing(16-spacing);
    m_buttonsHLayout->setContentsMargins(24, 0, 24, 0);

    m_mainVLayout->setSpacing(0);
    m_mainVLayout->addSpacing(16);
    m_mainVLayout->addLayout(m_settingsFormLayout);
    //m_mainVLayout->addStretch();
    m_mainVLayout->addSpacing(16);
    m_mainVLayout->addLayout(m_buttonsHLayout);
    m_mainVLayout->addSpacing(24);
    m_mainVLayout->setContentsMargins(0, 0, 0, 0);


    setLayout(m_mainVLayout);
}

void ScanSettingsWidget::initConnect()
{
    connect(m_scanButton, &QPushButton::clicked, this, &ScanSettingsWidget::scanButtonClickedSlot);

    connect(m_deviceComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::deviceCurrentTextChangedSlot);
    connect(m_pageNumberComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::pageNumberCurrentTextChangedSlot);
    connect(m_timeComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::timeCurrentTextChangedSlot);
    connect(m_typeComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::typeCurrentTextChangedSlot);
    connect(m_colorComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::colorCurrentTextChangedSlot);
    connect(m_resolutionComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::resolutionCurrentTextChangedSlot);
    connect(m_sizeComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::sizeCurrentTextChangedSlot);
    connect(m_formatComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::formatCurrentTextChangedSlot);
    connect(m_saveNameEdit, &QLineEdit::textChanged, this, &ScanSettingsWidget::nameCurrentTextChangedSlot);
    connect(m_saveDirectoryButton, &QPushButton::clicked, this, &ScanSettingsWidget::selectSaveDirectorySlot);

    connect(m_sendMailButton, &QPushButton::clicked, this, &ScanSettingsWidget::sendMailButtonClickedSlot);
    connect(m_SaveAsButton, &QPushButton::clicked, this, &ScanSettingsWidget::saveAsButtonClickedSlot);
    connect(g_user_signal, &GlobalUserSignal::closeWindowSaveAsSignal, this, &ScanSettingsWidget::saveAsButtonClickedSlot);

    connect(kdk::kabase::Gsettings::getPoint(), &kdk::kabase::Gsettings::systemFontSizeChange, this, &ScanSettingsWidget::fontSizeChangedSlot);
//    connect(kdk::kabase::Gsettings::getPoint(), &kdk::kabase::Gsettings::systemThemeChange, this, &ScanSettingsWidget::initTheme);

    connect(g_user_signal, &GlobalUserSignal::scanThreadFinishedSignal, this, &ScanSettingsWidget::updateSettingsStatusForEndScan);
    connect(g_user_signal, &GlobalUserSignal::stopOrFinishedScanSignal, this, &ScanSettingsWidget::updateSaveNameTextWhileStopOrFinishedScan);

}

void ScanSettingsWidget::showTimeRow()
{
    int countRow = m_settingsFormLayout->rowCount();
    KyInfo() << "countRow = " << countRow;

    if (countRow == 12) {
        if (m_timeLabel) {
            m_timeLabel = new QLabel;
            setLabelAttributes(m_timeLabel, tr("Time"));
        }
        if (m_timeComboBox) {
            m_timeComboBox = new QComboBox;
//            m_timeComboBox->setFixedSize(ScanSettingsWidgetComboboxSize);
            connect(m_timeComboBox, &QComboBox::currentTextChanged, this, &ScanSettingsWidget::timeCurrentTextChangedSlot);
        }
        m_settingsFormLayout->insertRow(4, m_timeLabel, m_timeComboBox);

        updateTimeSettings();
    }
}

void ScanSettingsWidget::hideTimeRow()
{
    int countRow = m_settingsFormLayout->rowCount();
    KyInfo() << "countRow = " << countRow;

    if (countRow == 13) {
        m_settingsFormLayout->removeRow(4);
    }
}

void ScanSettingsWidget::updateScanButtonSettings()
{
    bool saneStatus = g_sane_object->getSaneStatus();
    m_scanButton->setEnabled(saneStatus);
}

void ScanSettingsWidget::updateDeviceSettings()
{
    QStringList deviceStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    if (! saneStatus) {
        /// 1. Not find any scan device
        /// 2. This device open false

        deviceStringList = g_sane_object->getSaneNames();
        if (deviceStringList.count() > 1) {
            // Not only one scan device, just this scanner open false, we need set all sane names
        } else {
            deviceStringList << tr("No available scanners");
        }
    } else {
        deviceStringList = g_sane_object->getSaneNames();
    }

    m_deviceComboBox->setEnabled(true);

    KyInfo() << "Devices :  " << deviceStringList;
    setComboboxAttributes(m_deviceComboBox, deviceStringList);
}

/**
 * @brief ScanSettingsWidget::updatePageNumberSettings
 * Single page:
 *      user choose it, not get from scanners,
 *      only scan once.
 *
 * Multiple page:
 *      user choose it, not get from scanners
 *      scan lots of times, only stop by clicked scan button again or not pages in ADF scanners
 */
void ScanSettingsWidget::updatePageNumberSettings()
{
    QStringList pageNumberStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    pageNumberStringList << tr("Single") << tr("Multiple");

    KyInfo() << "PageNumber :  " << pageNumberStringList;

    m_pageNumberComboBox->setEnabled(saneStatus);
    setComboboxAttributes(m_pageNumberComboBox, pageNumberStringList);
}

void ScanSettingsWidget::updateTimeSettings()
{
    QStringList timerStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    timerStringList << tr("3s") << tr("5s") << tr("7s") << tr("10s") << tr("15s");

    KyInfo() << "Timer :  " << timerStringList;

    m_timeComboBox->setEnabled(saneStatus);
    setComboboxAttributes(m_timeComboBox, timerStringList);

}

void ScanSettingsWidget::updateTypeSettings()
{
    QStringList typeStringList;

    bool saneStatus = g_sane_object->getSaneStatus();


    if (! saneStatus) {
        typeStringList << tr("Flatbed") << tr("ADF");
    } else {
        typeStringList = g_sane_object->getSaneTypes();
    }

    KyInfo() << "Type :  " << typeStringList;

    m_typeComboBox->setEnabled(saneStatus);
    if (g_sane_object->haveSourceFlag == 0) {
        // For ESCL scanner device, which not get source type parameter
        //m_typeComboBox->setEnabled(false);
    }
    setComboboxAttributes(m_typeComboBox, typeStringList);
}

void ScanSettingsWidget::updateColorSettings()
{
    QStringList colorStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    if (! saneStatus) {
        colorStringList << tr("Gray") << tr("Color") << tr("Lineart");
    } else {
        colorStringList = g_sane_object->getSaneColors();
    }


    // Avoid segmentation fault
    if (! colorStringList.isEmpty()) {
        KyInfo() << "Color: " << colorStringList;
    }

    m_colorComboBox->setEnabled(saneStatus);
    setComboboxAttributes(m_colorComboBox, colorStringList);
}

void ScanSettingsWidget::updateResolutionSettings(bool DetectedDevices)
{
    QStringList resolutionStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    if (! saneStatus) {
        resolutionStringList << tr("75 dpi") << tr("100 dpi") << tr("150 dpi");
    } else {
        resolutionStringList = g_sane_object->getSaneResolutions();
    }

    if (! resolutionStringList.isEmpty()) {
        KyInfo() << "resolution: " << resolutionStringList;

        m_resolutionComboBox->setEnabled(saneStatus);
        setComboboxAttributes(m_resolutionComboBox, resolutionStringList);
    } else {
        /// Cannot get resolution from scanner: FUJITSU fi-7140
        resolutionStringList.clear();
        KyInfo() << "resolution is empty! ";
        m_resolutionComboBox->setEnabled(false);

        if (DetectedDevices) {
            QString msg = tr("Resolution is empty!");
            warnMsg(msg);
        }
    }

}


void ScanSettingsWidget::updateSizeSettings()
{
    QStringList sizeStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    if (! saneStatus) {
        sizeStringList << tr("A4") << tr("A5");
    } else {
        sizeStringList = g_sane_object->getSaneSizes();
    }

    m_sizeComboBox->setEnabled(saneStatus);
    setComboboxAttributes(m_sizeComboBox, sizeStringList);
}

/**
 * @brief ScanSettingsWidget::updateFormatSettings
 * format is choosed by user, so we need set it default.
 */
void ScanSettingsWidget::updateFormatSettings()
{
    QStringList formatStringList;

    bool saneStatus = g_sane_object->getSaneStatus();

    formatStringList << "jpg" << "png" << "pdf" << "bmp";

    m_formatComboBox->setEnabled(saneStatus);
    setComboboxAttributes(m_formatComboBox, formatStringList);

}

/**
 * @brief ScanSettingsWidget::updateSaveNameTextSettings
 * Consider multiple scan, so we set nameEdit timestamp.
 */
void ScanSettingsWidget::updateSaveNameTextSettings()
{
    bool saneStatus = g_sane_object->getSaneStatus();

    QDateTime currentTime = QDateTime::currentDateTime();
    KyInfo() << "current time: " <<currentTime;

    QString currentTimeString = currentTime.toString("yyyy_MM_dd-hh_mm_ss_zzz");
    KyInfo() << currentTimeString;

    m_saveNameEdit->setText(currentTimeString);
    m_saveNameEdit->setAlignment(Qt::AlignLeft);

    m_saveNameEdit->setEnabled(saneStatus);
}

void ScanSettingsWidget::updateSaveNameTextWhileStopOrFinishedScan()
{
    QString nowName = g_sane_object->nowSaveName;
    KyInfo() << "current nowName: " << nowName;

    m_saveNameEdit->setText(nowName);

}

void ScanSettingsWidget::updateSaveDirectorySettings()
{
    bool saneStatus = g_sane_object->getSaneStatus();
    m_saveDirectoryButton->setEnabled(saneStatus);
}

void ScanSettingsWidget::updateSendMailSettings()
{
    bool saneStatus = g_sane_object->getSaneStatus();
    m_sendMailButton->setEnabled(saneStatus);
}

void ScanSettingsWidget::updateSaveAsSettings()
{
    bool saneStatus = g_sane_object->getSaneStatus();
    m_SaveAsButton->setEnabled(saneStatus);
}

void ScanSettingsWidget::updateSaveAsTextStore()
{
    m_SaveAsButton->setText(tr("Store text"));

    KyInfo() << "m_SaveAsButton text: " << m_SaveAsButton->text();
}

void ScanSettingsWidget::updateSaveAsTextRecover()
{
    KyInfo() << "m_SaveAsButton text: " << m_SaveAsButton->text();
    m_SaveAsButton->setText(tr("Save as"));
}


/**
 * @brief ScanSettingsWidget::updateSettings
 * update scan settings while detect and open the default scanner successfully
 */
void ScanSettingsWidget::updateSettingsForDetectDevices()
{
    updateScanButtonSettings();
    updateDeviceSettings();
    updatePageNumberSettings();
    updateTimeSettings();
    updateTypeSettings();
    updateColorSettings();
    updateResolutionSettings(true);
    updateSizeSettings();
    updateFormatSettings();
    updateSaveNameTextSettings();
    updateSaveDirectorySettings();

    m_sendMailButton->setEnabled(false);
    m_SaveAsButton->setEnabled(false);
}

void ScanSettingsWidget::updateSettingsForSwitchDevices()
{
    updateScanButtonSettings();
    updatePageNumberSettings();
    updateTypeSettings();
    updateColorSettings();
    updateResolutionSettings(true);
    updateSizeSettings();
    updateFormatSettings();
    updateSaveNameTextSettings();
    updateSaveDirectorySettings();

//    updateSendMailSettings();
//    updateSaveAsSettings();

    m_sendMailButton->setEnabled(false);
    m_SaveAsButton->setEnabled(false);
}

void ScanSettingsWidget::updateSettingsStatusForStartScan()
{
    m_scanButton->setEnabled(false);
//    m_deviceComboBox->setEnabled(false);
    m_pageNumberComboBox->setEnabled(false);


    int countRow = m_settingsFormLayout->rowCount();
    if (countRow == 13) {
        m_timeComboBox->setEnabled(false);
    }
    m_typeComboBox->setEnabled(false);
    m_colorComboBox->setEnabled(false);
    m_resolutionComboBox->setEnabled(false);
    m_sizeComboBox->setEnabled(false);
    m_formatComboBox->setEnabled(false);
    m_saveNameEdit->setEnabled(false);
    m_saveDirectoryButton->setEnabled(false);
    m_sendMailButton->setEnabled(false);
    m_SaveAsButton->setEnabled(false);
}

void ScanSettingsWidget::updateSettingsStatusForEndScan(int saneStatus)
{
    KyInfo() << "status: " << saneStatus;
    m_deviceComboBox->setEnabled(true);

    if (saneStatus == SANE_STATUS_NO_DOCS
            || saneStatus == SANE_STATUS_DEVICE_BUSY
            || saneStatus == SANE_STATUS_CANCELLED) {

        m_scanButton->setEnabled(true);
        m_pageNumberComboBox->setEnabled(true);
        int countRow = m_settingsFormLayout->rowCount();
        if (countRow == 13) {
            m_timeComboBox->setEnabled(true);
        }
        m_typeComboBox->setEnabled(true);
        m_colorComboBox->setEnabled(true);
        m_resolutionComboBox->setEnabled(true);
        m_sizeComboBox->setEnabled(true);
        m_formatComboBox->setEnabled(true);
        m_saveNameEdit->setEnabled(true);
        m_saveDirectoryButton->setEnabled(true);
        m_sendMailButton->setEnabled(true);
        m_SaveAsButton->setEnabled(true);

    } else if (saneStatus == SANE_STATUS_INVAL) {
        m_scanButton->setEnabled(true);
        m_pageNumberComboBox->setEnabled(true);

        int countRow = m_settingsFormLayout->rowCount();
        if (countRow == 13) {
            m_timeComboBox->setEnabled(true);
        }

        m_typeComboBox->setEnabled(true);
        m_colorComboBox->setEnabled(true);
        m_resolutionComboBox->setEnabled(true);
        m_sizeComboBox->setEnabled(true);
        m_formatComboBox->setEnabled(true);
        m_saveNameEdit->setEnabled(true);
        m_saveDirectoryButton->setEnabled(true);

        m_sendMailButton->setEnabled(false);
        m_SaveAsButton->setEnabled(false);
    } else {
        bool saneStatusBool = true;
        if (saneStatus == 0) {
            saneStatusBool = true;
        } else {
            saneStatusBool = false;
        }
        m_scanButton->setEnabled(saneStatusBool);
        m_pageNumberComboBox->setEnabled(saneStatusBool);

        int countRow = m_settingsFormLayout->rowCount();
        if (countRow == 13) {
            m_timeComboBox->setEnabled(saneStatusBool);
        }
        m_typeComboBox->setEnabled(saneStatusBool);
        m_colorComboBox->setEnabled(saneStatusBool);
        m_resolutionComboBox->setEnabled(saneStatusBool);
        m_sizeComboBox->setEnabled(saneStatusBool);
        m_formatComboBox->setEnabled(saneStatusBool);
        m_saveNameEdit->setEnabled(saneStatusBool);
        m_saveDirectoryButton->setEnabled(saneStatusBool);
        m_sendMailButton->setEnabled(saneStatusBool);
        m_SaveAsButton->setEnabled(saneStatusBool);
    }
}

void ScanSettingsWidget::setLabelAttributes(QLabel *label, const QString &text, int labelHeight)
{
    label->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    label->setFixedWidth(ScanSettingsLabelWidth);
    label->setText(text);

#if 1
    int elideWidth = ScanSettingsLabelElideWidth;

    QFontMetrics fontMetrics(label->font());

    if (fontMetrics.width(text) > label->width()) {
        QString elideText = QFontMetrics(label->font()).elidedText(text, Qt::ElideRight, elideWidth);
        label->setText(elideText);
        label->setToolTip(text);
    } else {
        label->setText(text);
        label->setToolTip("");
    }
#endif
}


void ScanSettingsWidget::setSaveButtonAttributes(QPushButton *button, const QString &text, int elideWidth)
{
    button->setStyleSheet("QPushButton{text-align:left;}");
    QFontMetrics fontMetrics(button->font());

    if (fontMetrics.width(text) > elideWidth) {
        QString elideText = QFontMetrics(button->font()).elidedText(text, Qt::ElideRight, elideWidth);
        button->setText(elideText);
        button->setToolTip(text);
    } else {
        button->setText(text);
        button->setToolTip("");
    }

}

void ScanSettingsWidget::setSaveButtonLabelAttributes(QLabel *label, const QString &text, int elideWidth)
{
    QFontMetrics fontMetrics(label->font());

    if (fontMetrics.width(text) > elideWidth) {
        QString elideText = QFontMetrics(label->font()).elidedText(text, Qt::ElideRight, elideWidth);
        label->setText(elideText);
        label->setToolTip(text);
    } else {
        label->setText(text);
        label->setToolTip("");
    }

}

void ScanSettingsWidget::setComboboxAttributes(QComboBox *combobox, QStringList strList)
{
    QStandardItemModel *model = new QStandardItemModel();
    int maxTooltipTextLength = 20;

    combobox->clear();

    for (int i = 0; i < strList.size(); ++i) {
        QStandardItem *item = new QStandardItem(strList.at(i));;
        int curTextLen = strList.at(i).length();

        if ( curTextLen >= maxTooltipTextLength) {
            item->setToolTip(strList.at(i));
        } else {
            item->setToolTip("");
        }

        model->appendRow(item);
    }
    combobox->setModel(model);
    combobox->setInsertPolicy(QComboBox::NoInsert);
    combobox->setFocusPolicy(Qt::ClickFocus);
}

void ScanSettingsWidget::setNameEditTooltip()
{
    QString str = m_saveNameEdit->text ();
    m_saveNameEdit->setToolTip (str);
}
void ScanSettingsWidget::warnMsg(QString msg)
{
    QMessageBox *msgBox = new QMessageBox();

    msgBox->setText(msg);
    msgBox->setIcon(QMessageBox::Warning);
//    msgBox->setWindowIcon(QIcon::fromTheme("kylin-scanner"));
    msgBox->setWindowTitle(tr("Scanner"));
    msgBox->setStandardButtons(QMessageBox::Yes);
    msgBox->setContextMenuPolicy(Qt::NoContextMenu);
    msgBox->button(QMessageBox::Yes)->setText(tr("Yes"));

    QWidget *widget = nullptr;
    QWidgetList widgetList = QApplication::allWidgets();
    for (int i=0; i<widgetList.length(); ++i) {
        if (widgetList.at(i)->objectName() == "MainWindow") {
            widget = widgetList.at(i);
        }
    }
    if (widget) {
        msgBox->setParent(widget);
        QRect rect = widget->geometry();
        int x = rect.x() + rect.width()/2 - msgBox->width()/2;
        int y = rect.y() + rect.height()/2 - msgBox->height()/2;
        msgBox->move(x,y);
    }

    msgBox->exec();
}
